<?php

/**
 * @file
 * Handles incoming requests from Sphinx indexer to generate XMLPipe stream.
 * This script should be executed from the PHP CLI.
 */

/**
 * Parse script arguments.
 */
$bootstrap_arguments = array(
  'host' => '',
  'root' => '',
  'uid'  => 0,
  'lang' => '',
);
if (!empty($_SERVER['argv'])) {
  $argc = count($_SERVER['argv']);
  for ($index = 1; $index < $argc; $index++) {
    if (!preg_match('`^--(.*?)=(.*)$`', $_SERVER['argv'][$index], $matches)) {
      print "\nERROR: Invalid argument: {$_SERVER['argv'][$index]}, value is missing.\n";
      exit(1);
    }
    if (isset($bootstrap_arguments[$matches[1]])) {
      $bootstrap_arguments[$matches[1]] = $matches[2];
    }
    else {
      // Other arguments are passed via $_GET for additional processing.
      $_GET[$matches[1]] = $matches[2];
    }
  }
}

// Check required arguments.
if (empty($bootstrap_arguments['host']) || empty($bootstrap_arguments['root'])) {
  print "\nERROR: Unable to locate site to bootstrap.\n";
  exit(1);
}

// Check Drupal root path.
if (!file_exists($bootstrap_arguments['root'] . '/includes/bootstrap.inc')) {
  print "\nERROR: Unable to locate Drupal installation at '{$bootstrap_arguments['root']}'.\n";
  exit(1);
}

// Move current working directory to Drupal root.
chdir($bootstrap_arguments['root']);

// Setup the proper environment for Drupal.
$_SERVER['HTTP_HOST'] = $_SERVER['SERVER_NAME'] = $bootstrap_arguments['host'];
$_SERVER['REMOTE_ADDR'] = '127.0.0.1';
$_SERVER['SERVER_SOFTWARE'] = 'PHP CLI';
$_SERVER['REQUEST_METHOD'] = 'GET';
$_SERVER['QUERY_STRING'] = '';
$_SERVER['PHP_SELF'] = $_SERVER['REQUEST_URI'] = '/';
$_SERVER['SCRIPT_NAME'] = '/xmlpipe.php';
if (!isset($_GET['q'])) {
  $_GET['q'] = '';
}

// Load and boot Drupal.
include_once './includes/bootstrap.inc';
drupal_bootstrap(DRUPAL_BOOTSTRAP_CONFIGURATION);
if (empty($GLOBALS['db_url'])) {
  print "\nERROR: Unable to locate the host '{$bootstrap_arguments['host']}' in the specified Drupal installation.\n";
  exit(1);
}
drupal_bootstrap(DRUPAL_BOOTSTRAP_DATABASE);
if (empty($GLOBALS['active_db'])) {
  print "\nERROR: Could not connect to the database.\n";
  exit(1);
}
drupal_bootstrap(DRUPAL_BOOTSTRAP_FULL);

// Register shutdown function to clean up user session.
function sphinxsearch_xmlpipe_shutdown() {
  db_query("DELETE FROM {sessions} WHERE hostname = '127.0.0.1'");
}
register_shutdown_function('sphinxsearch_xmlpipe_shutdown');

// Check maintenance mode.
if (variable_get('maintenance_mode', 0)) {
  watchdog('sphinx', 'XMLPipe could not run because the site is in maintenance mode.', array(), WATCHDOG_WARNING);
  print "\nERROR: XMLPipe could not run because the site is in maintenance mode.\n";
  exit(1);
}

// Make sure the module is really enabled.
if (!module_exists('sphinxsearch')) {
  watchdog('sphinx', 'XMLPipe could not run because the module is not enabled.', array(), WATCHDOG_WARNING);
  print "\nERROR: XMLPipe could not run because the module is not enabled.\n";
  exit(1);
}

// Enable PHP strict error reporting.
error_reporting(E_ALL | E_STRICT);

// Setup current user and language.
if (!empty($bootstrap_arguments['uid'])) {
  if ($GLOBALS['user']->uid != $bootstrap_arguments['uid']) {
    if (!($account = user_load($bootstrap_arguments['uid']))) {
      print "\nERROR: User {$bootstrap_arguments['uid']} does not exist.\n";
      exit(1);
    }
    $GLOBALS['user'] = $account;
  }
}

// Setup current language.
if (!empty($bootstrap_arguments['lang'])) {
  if ($GLOBALS['language']->language != $bootstrap_arguments['lang']) {
    $languages = language_list();
    if (!isset($languages[$bootstrap_arguments['lang']])) {
      print "\nERROR: Language code '{$bootstrap_arguments['lang']}' not supported.\n";
      exit(1);
    }
    $GLOBALS['language'] = $languages[$bootstrap_arguments['lang']];
  }
}

// Load XMLPipe generator code.
$module_path = drupal_get_path('module', 'sphinxsearch');
require_once($module_path .'/sphinxsearch.common.inc');
require_once($module_path .'/sphinxsearch.xmlpipe.inc');

// Launch XMLPipe generator and exit.
sphinxsearch_xmlpipe();
exit(0);
